/*
  tfkatd - An internet talker
  Copyright (C) 2001 Peter Howkins

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Email: tfkatd@calisto.cx

 */
/* Ansi includes */
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* socket includes */
#include "netinet/in.h"
#include "sys/ioctl.h"
#include "sys/socket.h"
#include "sys/time.h"
#include "sys/types.h"
#include "unistd.h"

#include "arpa/telnet.h"


#define STATE_NONE  0
#define STATE_LOGIN 1
#define STATE_PLAY  2

#define PORT   1067
#define MAXCON 20

typedef struct {
  int iSFD;
  int iState;
  char sName[12];
  char sLine[256];       /* one line of input */
} player;

/* globals */
player aPlayers[MAXCON]; /* all the player data */

fd_set readfds, creadfds;

/* function prototypes */
void send_to_player(int iPlayer, const char *sString);
void send_to_all(char *sString);
void close_connection(int p);
void do_login(int p);
void do_play(int p);




/* --- socket funcs ----------------------------------- */
void send_to_player(int iPlayer, const char *sString)
{
  int iCheck;

  /* bail out if socket if not connected to anyone */
  if(aPlayers[iPlayer].iState == STATE_NONE) {
    fprintf(stderr, "In a send_to_player() call when in STATE_NONE on %d", aPlayers[iPlayer].iSFD);
    return;
  }
  
  iCheck = send(aPlayers[iPlayer].iSFD, sString, strlen(sString), MSG_NOSIGNAL | MSG_DONTWAIT);

  if(iCheck == -1) {
    close_connection(iPlayer);
  }
}

/* send to all playing chars, (not those at login prompt) */
void send_to_all(char *sString)
{
  int iLoop;

  for(iLoop=0; iLoop<MAXCON; iLoop++) {
    if(aPlayers[iLoop].iState == STATE_PLAY) {
      send_to_player(iLoop, sString);
    }
  }   
}

int get_line_from_socket(int p)
{
  int iSize, iCount = 0;
  char cChar;
  char *sBuffer = aPlayers[p].sLine;
  size_t szBufsize = sizeof(aPlayers[p].sLine);

  do {
    iSize = recv(aPlayers[p].iSFD, &cChar, 1, MSG_NOSIGNAL | MSG_DONTWAIT);
    switch (iSize) {
      case 0:
      case -1:
        /* if logged in, inform other players of their exit */ 
        if(aPlayers[p].iState == STATE_PLAY) {
          char sTemp[256];
          snprintf(sTemp, sizeof(sTemp), "+++ %s disappears suddenly\r\n", aPlayers[p].sName);
          send_to_all(sTemp);
         }
        close_connection(p);
        *sBuffer = '\0';
        return -1;
    }
    /* Character read OK */
    *sBuffer++ = cChar;
    iCount++;
  } while (cChar != '\n' && iCount < szBufsize);
  if (cChar == '\n') {
    /* We read all the way up to a newline */
    /* handle LF and CRLF line termination */
    if(*(sBuffer - 2) == '\r') {
      *(sBuffer - 2) = '\0';
      return iCount - 2;
    } else {
      *(sBuffer - 1) = '\0';
      return iCount - 1;
    }
  } else {
    /* We ran out of buffer: terminate and return as much as possible */
    *(sBuffer - 1) = '\0';
    return iCount - 1;
  }
}

/* --- player init and close funcs ------------- */
void close_connection(int iPlayer)
{
  aPlayers[iPlayer].iState = STATE_NONE;      /* set the socket to unconected state */
  close(aPlayers[iPlayer].iSFD);

  /* remove sfd from fd sets */
  FD_CLR(aPlayers[iPlayer].iSFD, &readfds);

  fprintf(stderr, "Losing %d\n", iPlayer);
}

int get_free_player_space(void)
{
  int iLoop;
  
  for(iLoop=0; iLoop<MAXCON; iLoop++) {
    if(aPlayers[iLoop].iState == STATE_NONE) {
      return iLoop;
    }
  }
  return -1; /* to signify there is no more space */
}

void new_connection(int iSFD)
{
  int iIndex;

  /* find some space for the char */
  iIndex = get_free_player_space();
  if(iIndex == -1) {
    close(iSFD);
    return;
  }
  aPlayers[iIndex].iSFD = iSFD;  

  /* Add sfd to file descriptor sets */
  FD_SET(iSFD, &readfds);

  aPlayers[iIndex].iState = STATE_LOGIN;
  
  send_to_player(iIndex, "Login: ");
  
  fprintf(stderr, "New connection on %d\n", iIndex);
}

/* --- State handlind functions ----------------*/
void do_login(int iPlayer)
{
  int i;
  char *pcCheck;
  char sTemp[256];

  if(strcmp(aPlayers[iPlayer].sLine, "") == 0) {
    send_to_player(iPlayer, "That's a bit short. Try something longer\r\nLogin: ");
    return;
  }

  /* check to see if anyone else is using that name */
  for(i=0;i<MAXCON;i++) {
    if(aPlayers[i].iState == STATE_PLAY) {
      if(strcasecmp(aPlayers[iPlayer].sLine, aPlayers[i].sName) == 0) {
        send_to_player(iPlayer, "Sorry someone else is using that name. Try another\r\nLogin: ");
        return;
      }
    }
  }

  /* check that the name is alpha numeric */
  pcCheck = aPlayers[iPlayer].sLine;
  while(*pcCheck) {
    if(!isalnum(*pcCheck)) {
      send_to_player(iPlayer, "Sorry, names must only contain alphanumeric characters\r\nLogin: ");
      return;
    }
    pcCheck++;
  }

  /* they have entered something, make it there name */
  /* *************** FIX (buffer safety) ***********/
  strcpy(aPlayers[iPlayer].sName, aPlayers[iPlayer].sLine);

  aPlayers[iPlayer].iState = STATE_PLAY;

  snprintf(sTemp, sizeof(sTemp), "+++ %s has logged in\r\n", aPlayers[iPlayer].sName);
  send_to_all(sTemp);

}

void say(int iPlayer, char *sString)
{
  char sTemp[256];
  snprintf(sTemp, sizeof(sTemp), "%12s : %s\r\n", aPlayers[iPlayer].sName, sString);

  send_to_all(sTemp);
}

void emote(int iPlayer, const char *sString)
{
  char sTemp[256];
  snprintf(sTemp, sizeof(sTemp), "%s %s\r\n", aPlayers[iPlayer].sName, sString);

  send_to_all(sTemp);
}

void command(int iPlayer, const char *sCommand)
{
  if(strcmp(sCommand, "quit") == 0) {
    char sBuf[256];
    snprintf(sBuf, sizeof(sBuf), "+++ %s logs off\r\n", aPlayers[iPlayer].sName);
    send_to_all(sBuf);
    close_connection(iPlayer);
  } else if(strcmp(sCommand, "who") == 0) {
    int iLoop;
    send_to_player(iPlayer, "+++ ");
    for(iLoop=0; iLoop<MAXCON; iLoop++) {
      if(aPlayers[iLoop].iState == STATE_PLAY) {
        send_to_player(iPlayer, aPlayers[iLoop].sName);
        send_to_player(iPlayer, " ");
      }
    }
    send_to_player(iPlayer, "\r\n");
  } else if(strcmp(sCommand, "help") == 0) {
    send_to_player(iPlayer,
    "+++ TfkatD v0.02 GPL (c) 2001 Peter Howkins/Flibble\r\n"
    "+++\r\n"
    "+++ Commands\r\n"
    "+++  .who        : list who is connected\r\n"
    "+++  .quit       : log out\r\n"
    "+++  .help       : this file\r\n"
    "+++\r\n"
    "+++  ;<line>     : emote <line>\r\n"
    "+++  <line>      : say <line>\r\n");    
  } else {
    send_to_player(iPlayer, "+++ Command .");
    send_to_player(iPlayer, sCommand);
    send_to_player(iPlayer, " not known try .help\r\n");
  }
}

void do_play(int iPlayer)
{
  if(!strcmp(aPlayers[iPlayer].sLine, "")==0) {
    char *pcTemp = aPlayers[iPlayer].sLine;

    /* remove leading whitespace from line */
    while(*pcTemp && isspace(*pcTemp)) {
      pcTemp++;
    }

    if(*pcTemp == ';') {
      emote(iPlayer, pcTemp + 1);
    } else if(*pcTemp == '.') {
      command(iPlayer, pcTemp + 1);
    } else {
      say(iPlayer, pcTemp);
    }
  }
}

/* --- Initialisation funcs --------------------*/
void init(void)
{
  int iLoop;

  /* set all people to be not logged in yet */
  for(iLoop=0; iLoop<MAXCON; iLoop++) {
    aPlayers[iLoop].iState = STATE_NONE;
  }
}

int init_mother(int iPort)
{
  int iMotherSFD;
  struct sockaddr_in sai;
  int iOption = 1;

  /* Create mother description socket, and set socket options */
  iMotherSFD = socket(PF_INET, SOCK_STREAM, 0);
  if (iMotherSFD == -1) {
    fprintf(stderr, "Could not open mother descriptor socket: %s", strerror(errno));
    exit(EXIT_FAILURE);
  }

  /* Enable quick reuse of the port when we exit */
  if (setsockopt(iMotherSFD, SOL_SOCKET, SO_REUSEADDR,
                 (char *) &iOption, sizeof(iOption)) == -1) {
    fprintf(stderr,"Could not set mother descriptor reuse options: %s", strerror(errno));
  }

  /* Set non blocking */
  if (ioctl(iMotherSFD, FIONBIO, &iOption) == -1) {
    fprintf(stderr, "Could not set mother descriptor to non blocking: %s", strerror(errno));
    close(iMotherSFD);
    exit(EXIT_FAILURE);
  }

  sai.sin_family      = AF_INET;
  sai.sin_port        = htons(iPort);
  sai.sin_addr.s_addr = htonl(INADDR_ANY);

  /* Bind mother descriptor socket */
  if (bind(iMotherSFD, (struct sockaddr *) &sai, sizeof(sai)) < 0) {
    fprintf(stderr, "Could not bind socket: %s", strerror(errno));
    close(iMotherSFD);
    exit(EXIT_FAILURE);
  }

  /* Set socket to listen */
  if (listen(iMotherSFD, 3) ) {
    fprintf(stderr, "Could not listen() on socket: %s", strerror(errno));
    close(iMotherSFD);
    exit(EXIT_FAILURE);
  }
  return iMotherSFD;
}

int main(void)
{
  int iMotherSFD;  /* the mother socket file desciptor */

  /* LIBC 5 */
/*  int len = sizeof(struct sockaddr_in); */ /* for the accept call later */
  /* GLIBC 2 */
  socklen_t len = sizeof(struct sockaddr_in);  /* for the accept call later */
  struct timeval tv;

  /* initialise some gubins */
  init();
  iMotherSFD = init_mother(PORT);

  fprintf(stderr, "Tfkatd booted on port %d\n", PORT);

  /* the main loop */
  for(;;) {
    int iNewConnectionSFD;
    int iRetval;
    struct sockaddr_in sin;

    /* See if there's a new connection ready */
    iNewConnectionSFD = accept(iMotherSFD, (struct sockaddr *) &sin, &len);
    if (iNewConnectionSFD != -1) {
      new_connection(iNewConnectionSFD);
    }

    /* Use select() to see if there is any data to be read on any sfd, set
       time struct for select, this below should make it loop 5 times a 
       second (0.2 seconds) and copy the fd sets so they don't get corrupted */
    tv.tv_sec = 0; tv.tv_usec = 200;
    creadfds = readfds;
    iRetval = select(FD_SETSIZE, &creadfds, NULL, NULL, &tv);

    if(iRetval != -1) {
      int iLoop;

      for(iLoop=0; iLoop<MAXCON; iLoop++) {
        if(FD_ISSET(aPlayers[iLoop].iSFD, &creadfds)) {
          if(get_line_from_socket(iLoop) != -1) {
            /* ignore lines with telnet command codes */
            if(aPlayers[iLoop].sLine[0] != (char) IAC) { 
              switch(aPlayers[iLoop].iState) {
                case STATE_LOGIN:
                  do_login(iLoop);
                  break;
                case STATE_PLAY:
                  do_play(iLoop);
                  break; 
              }
            }
            /* Ensure that sfd is checked next time 
               (only if getlinefromsock != -1)*/
            FD_SET(aPlayers[iLoop].iSFD, &readfds);  
          }
        }
      }
    }
  } /* end of main loop */

  return 0;
}
