/*
  tfkatdSSL - An internet talker with SSL transport layer
  Copyright (C) 2001 Peter Howkins

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Email: tfkatd@calisto.cx

 */
/* Ansi includes */
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* socket includes */
#include "netinet/in.h"
#include "sys/ioctl.h"
#include "sys/socket.h"
#include "sys/time.h"
#include "sys/types.h"
#include "unistd.h"

#include "arpa/telnet.h"

#include "openssl/bio.h"
#include "openssl/err.h"
#include "openssl/ssl.h"

#define STATE_NONE  0
#define STATE_LOGIN 1
#define STATE_PLAY  2

#define PORT   1068
#define MAXCON 20

typedef struct {
  SSL *hSSL;
  int iSFD;
  int iState;
  char sName[12];
  char sLine[256];       /* one line of input */
} player;

/* globals */
/* In most talkers this data is in a linked list of arbitrary length */
/* it's in an array purely for simplicity */
player aPlayers[MAXCON]; /* all the player data */

fd_set readfds, creadfds;

/* function prototypes */
void send_to_player(int iPlayer, const char *sString);
void send_to_all(char *sString);
void close_connection(int p);
void do_login(int p);
void do_play(int p);


SSL_CTX *hSSLContext = NULL;
BIO *hError = NULL;

/* --- socket funcs ----------------------------------- */
void send_to_player(int iPlayer, const char *sString)
{
  int iCheck;

  assert(iPlayer < MAXCON);
  assert(sString);

//  fprintf(stderr, "send_to_player: %d %s\n", iPlayer, sString);

  /* bail out if socket if not connected to anyone */
  if(aPlayers[iPlayer].iState == STATE_NONE) {
    fprintf(stderr, "In send_to_player() call when in STATE_NONE on %d", aPlayers[iPlayer].iSFD);
    return;
  }
  
  iCheck = SSL_write(aPlayers[iPlayer].hSSL, sString, strlen(sString));

  if(iCheck <= 0) {
    close_connection(iPlayer);
  }
}

/* send to all playing chars, (not those at login prompt) */
void send_to_all(char *sString)
{
  int iLoop;

  assert(sString);

  for(iLoop=0; iLoop<MAXCON; iLoop++) {
    if(aPlayers[iLoop].iState == STATE_PLAY) {
      send_to_player(iLoop, sString);
    }
  }   
}

int get_line_from_socket(int p)
{
  int iSize, iCount = 0;
  char cChar;
  char *sBuffer = aPlayers[p].sLine;
  size_t szBufsize = sizeof(aPlayers[p].sLine);

  assert(p < MAXCON);

  do {
    iSize = SSL_read(aPlayers[p].hSSL, &cChar, 1);
    switch (iSize) {
      case -1:
        /* if logged in, inform other players of their exit */ 
        if(aPlayers[p].iState == STATE_PLAY) {
          char sTemp[256];
          snprintf(sTemp, sizeof(sTemp), "+++ %s disappears suddenly\r\n", aPlayers[p].sName);
          send_to_all(sTemp);
         }
        close_connection(p);
        return -1;
/*      case -1:
        *sBuffer = '\0';
        return -1;
*/
    }
    /* Character read OK */
    *sBuffer++ = cChar;
    iCount++;
  } while (cChar != '\n' && iCount < szBufsize);
  if (cChar == '\n') {
    /* We read all the way up to a newline */
    /* handle LF and CRLF line termination */
    if(*(sBuffer - 2) == '\r') {
      *(sBuffer - 2) = '\0';
      return iCount - 2;
    } else {
      *(sBuffer - 1) = '\0';
      return iCount - 1;
    }
    return iCount - 1;
  } else {
    /* We ran out of buffer: terminate and return as much as possible */
    *(sBuffer - 1) = '\0';
    return iCount - 1;
  }
}

/* --- player init and close funcs ------------- */
void close_connection(int iPlayer)
{

  assert(iPlayer < MAXCON);

  aPlayers[iPlayer].iState = STATE_NONE;      /* set the socket to unconected state */
  close(aPlayers[iPlayer].iSFD);

  SSL_free(aPlayers[iPlayer].hSSL);

  /* remove sfd from fd sets */
  FD_CLR(aPlayers[iPlayer].iSFD, &readfds);

  fprintf(stderr, "Losing %d\n", iPlayer);
}

int get_free_player_space(void)
{
  int iLoop;
  
  for(iLoop=0; iLoop<MAXCON; iLoop++) {
    if(aPlayers[iLoop].iState == STATE_NONE) {
      return iLoop;
    }
  }
  return -1; /* to signify there is no more space */
}

void new_connection(int iSFD)
{
  int iIndex;
  int iOption = 1;
  int iCheck;

  assert(iSFD);

  /* find some space for the char */
  iIndex = get_free_player_space();
  if(iIndex == -1) {
    close(iSFD);
    return;
  }
  aPlayers[iIndex].iSFD = iSFD;  


  aPlayers[iIndex].hSSL = SSL_new(hSSLContext);
  if(NULL == aPlayers[iIndex].hSSL) {
    ERR_print_errors(hError);
    close(iSFD);
    return;
  }

  /* assign the Socket file descriptor to the SSL structure */
  if(SSL_set_fd(aPlayers[iIndex].hSSL, iSFD) == 0) {
    ERR_print_errors(hError);
    close(iSFD);
    return;
  }

  iCheck = SSL_accept(aPlayers[iIndex].hSSL);
  if(iCheck != 1) {
    ERR_print_errors(hError);  
    close(iSFD);
    return;
  }

  /* Set the socket to non-blocking */
  /* Only after handshaking is complete (this may lag the talker at times */
  if (ioctl(aPlayers[iIndex].iSFD, FIONBIO, &iOption) == EOF) {
    fprintf(stderr, "Could not set %d socket options: %s", iSFD, strerror(errno));
    close_connection(iIndex);
    return;
  }


  /* Add sfd to file descriptor sets */
  FD_SET(iSFD, &readfds);

  aPlayers[iIndex].iState = STATE_LOGIN;
  
  send_to_player(iIndex, "Login: ");
  
  fprintf(stderr, "New connection on %d\n", iIndex);
}

/* --- State handlind functions ----------------*/
void do_login(int iPlayer)
{
  int i;
  char *pcCheck;
  char sTemp[256];

  assert(iPlayer < MAXCON);

  if(strcmp(aPlayers[iPlayer].sLine, "") == 0) {
    send_to_player(iPlayer, "That's a bit short. Try something longer\r\nLogin: ");
    return;
  }

  /* check to see if anyone else is using that name */
  for(i=0;i<MAXCON;i++) {
    if(aPlayers[i].iState == STATE_PLAY) {
      if(strcasecmp(aPlayers[iPlayer].sLine, aPlayers[i].sName) == 0) {
        send_to_player(iPlayer, "Sorry someone else is using that name. Try another\r\nLogin: ");
        return;
      }
    }
  }

  /* check that the name is alpha numeric */
  pcCheck = aPlayers[iPlayer].sLine;
  while(*pcCheck) {
    if(!isalnum(*pcCheck)) {
      send_to_player(iPlayer, "Sorry, names must only contain alphanumeric characters\r\nLogin: ");
      return;
    }
    pcCheck++;
  }

  /* they have entered something, make it there name */
  snprintf(aPlayers[iPlayer].sName, sizeof(aPlayers[iPlayer].sName), "%s", aPlayers[iPlayer].sLine);

  aPlayers[iPlayer].iState = STATE_PLAY;

  snprintf(sTemp, sizeof(sTemp), "+++ %s has logged in\r\n", aPlayers[iPlayer].sName);
  send_to_all(sTemp);

}

void say(int iPlayer, char *sString)
{
  char sTemp[256];

  assert(iPlayer < MAXCON);
  assert(sString);

  snprintf(sTemp, sizeof(sTemp), "%12s : %s\r\n", aPlayers[iPlayer].sName, sString);

  send_to_all(sTemp);
}

void emote(int iPlayer, const char *sString)
{
  char sTemp[256];

  assert(iPlayer < MAXCON);
  assert(sString);

  snprintf(sTemp, sizeof(sTemp), "%s %s\r\n", aPlayers[iPlayer].sName, sString);

  send_to_all(sTemp);
}

void command(int iPlayer, const char *sCommand)
{
  assert(iPlayer < MAXCON);
  assert(sCommand);

  if(strcmp(sCommand, "quit") == 0) {
    char sBuf[256];
    snprintf(sBuf, sizeof(sBuf), "+++ %s logs off\r\n", aPlayers[iPlayer].sName);
    send_to_all(sBuf);
    close_connection(iPlayer);
  } else if(strcmp(sCommand, "who") == 0) {
    int iLoop;
    send_to_player(iPlayer, "+++ ");
    for(iLoop=0; iLoop<MAXCON; iLoop++) {
      if(aPlayers[iLoop].iState == STATE_PLAY) {
        send_to_player(iPlayer, aPlayers[iLoop].sName);
        send_to_player(iPlayer, " ");
      }
    }
    send_to_player(iPlayer, "\r\n");
  } else if(strcmp(sCommand, "help") == 0) {
    send_to_player(iPlayer,
    "+++ TfkatDSSL v0.02 GPL (c) 2001 Peter Howkins/Flibble\r\n"
    "+++\r\n"
    "+++ Commands\r\n"
    "+++  .who        : list who is connected\r\n"
    "+++  .quit       : log out\r\n"
    "+++  .help       : this file\r\n"
    "+++\r\n"
    "+++  ;<line>     : emote <line>\r\n"
    "+++  <line>      : say <line>\r\n");    
  } else {
    send_to_player(iPlayer, "+++ Command .");
    send_to_player(iPlayer, sCommand);
    send_to_player(iPlayer, " not known try .help\r\n");
  }
}

void do_play(int iPlayer)
{
  if(!strcmp(aPlayers[iPlayer].sLine, "")==0) {
    char *pcTemp = aPlayers[iPlayer].sLine;

    /* remove leading whitespace from line */
    while(*pcTemp && isspace(*pcTemp)) {
      pcTemp++;
    }

    if(*pcTemp == ';') {
      emote(iPlayer, pcTemp + 1);
    } else if(*pcTemp == '.') {
      command(iPlayer, pcTemp + 1);
    } else {
      say(iPlayer, pcTemp);
    }
  }
}

#define CERTF  "./server.crt"
#define KEYF   "./server.key"

/* --- Initialisation funcs --------------------*/
void init(void)
{
  int iLoop;

  /* set all people to be not logged in yet */
  for(iLoop=0; iLoop<MAXCON; iLoop++) {
    aPlayers[iLoop].iState = STATE_NONE;
  }

  /* Initialise the SSL library */  
	SSL_library_init();
	SSL_load_error_strings();

  /* Prepare the SSL error method */
  hError = BIO_new_fp(stderr, BIO_NOCLOSE);  
  if(!hError) {
    fprintf(stderr, "Failed to set up Error BIO\n");
    exit(EXIT_FAILURE);
  }

  /* Set up the global SSL context */
  hSSLContext = SSL_CTX_new(SSLv23_method());
  if(NULL == hSSLContext) {
    ERR_print_errors(hError);
    exit(EXIT_FAILURE);
  }


  if (SSL_CTX_use_certificate_file(hSSLContext, CERTF, SSL_FILETYPE_PEM) <= 0) {
    ERR_print_errors(hError);
    exit(EXIT_FAILURE);
  }
  if (SSL_CTX_use_PrivateKey_file(hSSLContext, KEYF, SSL_FILETYPE_PEM) <= 0) {
    ERR_print_errors(hError);
    exit(EXIT_FAILURE);
  }

  if (!SSL_CTX_check_private_key(hSSLContext)) {
    fprintf(stderr,"Private key does not match the certificate public key\n");
    exit(EXIT_FAILURE);
  }


  /* give it some options */
  SSL_CTX_set_options(hSSLContext, SSL_OP_ALL);
  SSL_CTX_set_cipher_list(hSSLContext, "RC4-MD5:EXP-RC4-MD5");  
//  SSL_CTX_set_cipher_list(hSSLContext, SSL_TXT_ALL);  
  SSL_CTX_set_verify(hSSLContext, SSL_VERIFY_NONE, NULL);
  
  
}

int init_mother(int iPort)
{
  int iMotherSFD;
  struct sockaddr_in sai;
  int iOption = 1;

  assert(iPort > 0);

  /* Create mother description socket, and set socket options */
  iMotherSFD = socket(PF_INET, SOCK_STREAM, 0);
  if (iMotherSFD == -1) {
    fprintf(stderr, "Could not open mother descriptor socket: %s", strerror(errno));
    exit(EXIT_FAILURE);
  }

  /* Enable quick reuse of the port when we exit */
  if (setsockopt(iMotherSFD, SOL_SOCKET, SO_REUSEADDR,
                 (char *) &iOption, sizeof(iOption)) == -1) {
    fprintf(stderr,"Could not set mother descriptor options1: %s", strerror(errno));
  }

  /* Set non blocking */
  if (ioctl(iMotherSFD, FIONBIO, &iOption) == -1) {
    fprintf(stderr, "Could not set mother descriptor options2: %s", strerror(errno));
    close(iMotherSFD);
    exit(EXIT_FAILURE);
  }

  sai.sin_family      = AF_INET;
  sai.sin_port        = htons(iPort);
  sai.sin_addr.s_addr = htonl(INADDR_ANY);

  /* Bind mother description socket */
  if (bind(iMotherSFD, (struct sockaddr *) &sai, sizeof(sai)) < 0) {
    fprintf(stderr, "Could not bind socket: %s", strerror(errno));
    close(iMotherSFD);
    exit(EXIT_FAILURE);
  }

  /* Set socket to listen */
  if (listen(iMotherSFD, 3) ) {
    fprintf(stderr, "Could not listen() on socket: %s", strerror(errno));
    close(iMotherSFD);
    exit(EXIT_FAILURE);
  }
  return iMotherSFD;
}

void init_signal(void)
{
  /* this should prevent writing to a closed socket quitting the pogram */
  signal(SIGPIPE, SIG_IGN);
}

int main(void)
{
  int iMotherSFD;  /* the mother socket file desciptor */

  /* LIBC 5 */
/*  int len = sizeof(struct sockaddr_in); */ /* for the accept call later */
  /* GLIBC 2 */
  socklen_t len = sizeof(struct sockaddr_in);  /* for the accept call later */
  struct timeval tv;

  /* initialise some gubins */
  init();
  iMotherSFD = init_mother(PORT);
  init_signal();

  fprintf(stderr, "Tfkatd booted on port %d\n", PORT);

  /* the main loop */
  for(;;) {
    int iNewConnectionSFD;
    int iRetval;
    struct sockaddr_in sin;

    /* See if there's a new connection ready */
    iNewConnectionSFD = accept(iMotherSFD, (struct sockaddr *) &sin, &len);
    if (iNewConnectionSFD != -1) {
      new_connection(iNewConnectionSFD);
    }

    /* Use select() to see if there is any data to be read on any sfd, set
       time struct for select, this below should make it loop 5 times a 
       second (0.2 seconds) and copy the fd sets so they don't get corrupted */
    tv.tv_sec = 1; tv.tv_usec = 0;
    creadfds = readfds;
    iRetval = select(FD_SETSIZE, &creadfds, NULL, NULL, &tv);


    if(iRetval != -1) {
      int iLoop;

      for(iLoop=0; iLoop<MAXCON; iLoop++) {
        if(FD_ISSET(aPlayers[iLoop].iSFD, &creadfds)) {
          if(get_line_from_socket(iLoop) != -1) {
            if(aPlayers[iLoop].sLine[0] != (char) IAC) { /* ignore lines with telnet command codes */
              switch(aPlayers[iLoop].iState) {
                case STATE_LOGIN:
                  do_login(iLoop);
                  break;
                case STATE_PLAY:
                  do_play(iLoop);
                  break; 
              }
            }
            /* Ensure that sfd is checked next time 
               (only if getlinefromsock != -1 *AND* they haven't been chucked
                off in the mean time)*/
            if(aPlayers[iLoop].iState != STATE_NONE) {
              FD_SET(aPlayers[iLoop].iSFD, &readfds);  
            }
          }
        }
      }
    }
  } /* end of main loop */

  return 0;
}
